/*
 * gpPmfDraw.c --
 *
 *
 */

#include "tkInt.h"
#include "tkCanvas.h"
#include "gpXInt.h"
#include "gpVisitor.h"

/*
 * Standard item types provided by Tk and GIPSI ones:
 */
 
extern Tk_ItemType tkArcType, tkBitmapType, tkImageType, tkLineType;
extern Tk_ItemType tkOvalType, tkPolygonType;
extern Tk_ItemType tkRectangleType, tkTextType, tkWindowType;
extern Tk_ItemType GpSymbolType, GpTextType;

static Tk_Uid noneUid = NULL;
static Tk_Uid firstUid = NULL;
static Tk_Uid lastUid = NULL;
static Tk_Uid bothUid = NULL;

static Tk_Uid arcUid = NULL;
static Tk_Uid chordUid = NULL;
static Tk_Uid pieSliceUid = NULL;

/*
 * A useful data structure.
 */

typedef struct {
    Tcl_Interp *interp;
    Tk_Canvas canvas;
    gpCS *cs;
    Tk_Uid tag;
} Gp_PmfData;

/*
 * Forward declarations for procedures defined in this file:
 */

SFUNC(PmfRenderLoop, int, (Gp_PmfData *data, Pmf *pmf, PmfEntity *pEnt));
SFUNC(Gp_PreprocessItem, void, (Gp_PmfData *data, Tk_ItemType *typePtr,
				Tk_Item **pitem));
SFUNC(Gp_PostprocessItem, void, (Tk_Canvas canvas, Tk_Item *itemPtr));

SFUNC(Gp_PmfAddPosition, void, (Gp_PmfData *data, double dVal, int isX,
		double *val));
SFUNC(Gp_PmfScaleWidth, void, (Gp_PmfData *data, double dVal, int *val));
SFUNC(Gp_PmfScaleDoubleWidth, void, (Gp_PmfData *data, double dVal,
				     double *val));

SFUNC(Gp_PmfDrawLine, void, (Gp_PmfData *data, Line *line,
		Attribute *override));
SFUNC(Gp_PmfDrawCircle, void, (Gp_PmfData *data, Circle *circle,
		Attribute *override));
SFUNC(Gp_PmfDrawText, void, (Gp_PmfData *data, Text *text,
		Attribute *override, excVoidList *pvl));
SFUNC(Gp_PmfDrawSymbol, void, (Gp_PmfData *data, Symbol *symbol,
		Attribute *override));
SFUNC(Gp_PmfDrawPolySymbol, void, (Gp_PmfData *data, PolySym *polysymbol,
		Attribute *override, excVoidList *pvl));
SFUNC(Gp_PmfDrawPolyLine, void, (Gp_PmfData *data, PolyLine *polyline,
		Attribute *override, excVoidList *pvl));
SFUNC(Gp_PmfDrawPolygon, void, (Gp_PmfData *data, Polygon *polygon,
		Attribute *override, excVoidList *pvl));

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawCmd --
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
Gp_PmfDrawCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    Pmf *pmf = NULL;
    gpAttList *patt = NULL;
    PmfEntity *pEnt = NULL;
    int iCode = TCL_ERROR;
    Gp_PmfData data;
    Tcl_CmdInfo canvasCmd, csCmd;
    char tag[20];
    static int id = 0;

    if (argc != 5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" .canvas file attributes csID\"", (char *)NULL);
	return TCL_ERROR;
    }

    if (!Tcl_GetCommandInfo(interp, argv[1], &canvasCmd)) {
	Tcl_AppendResult(interp, "invalid canvas \"", argv[1], "\"",
		(char *)NULL);
	goto error;
    }
    if (!fPmfOpen(&pmf, argv[2], READ, NULL)) {
        Tcl_AppendResult(interp, "bad plot file: \"", argv[2],
		(char *)NULL);
	return TCL_ERROR;
    }
    if (!fPmfCreate((Pmf **)&pEnt, ENTITY_PMF, NULL)) {
	Tcl_AppendResult(interp, "out of memory", (char *)NULL);
	goto error;
    }
    if ((patt = (gpAttList *)Gp_GetAttributes(interp, argv[3])) == NULL) {
        Tcl_AppendResult(interp, "bad attribute file: \"", argv[3],
		(char *)NULL);
	goto error;
    }
    if (!Tcl_GetCommandInfo(interp, argv[4], &csCmd)) {
	Tcl_AppendResult(interp, "invalid csID \"", argv[4], "\"",
		(char *)NULL);
	goto error;
    }

    pmf_set(pmf, PATTLIST, patt, NULL);
    sprintf(tag, "gplayer%d", id++);
    
    data.interp = interp;
    data.canvas = (Tk_Canvas) canvasCmd.clientData;
    data.cs = (gpCS *) csCmd.clientData;
    data.tag = Tk_GetUid(tag);

    iCode = PmfRenderLoop(&data, pmf, pEnt);

error:
    if (pmf) {	/* This will destroy patt so set it to nothing first. */ 
        pmf_set(pmf, PATTLIST, (gpAttList *)NULL, NULL);
	fPmfClose(pmf, KEEP);
    }
    if (patt) {
	Gp_FreeAttributes(patt);
    }
    if (pEnt) {
	DestroyPmf(pEnt);
    }
    if (iCode == TCL_OK) {
	Tcl_SetResult(interp, tag, TCL_VOLATILE);
    }
    return iCode;
}

/*
 *----------------------------------------------------------------------
 *
 * PmfRenderLoop --
 *
 *	This procedure will loop over the pmf file and draw the items.
 *
 * Results:
 *      If an error occurs the TCL_ERROR is returned with an error
 *	message in data->interp.result. Otherwise TCL_OK is returned,
 *
 * Side effects:
 *	A pmf file will eventually be read. Also, for large pmf files
 *	X Events will be handled to check if rendering should be halted.
 *
 *----------------------------------------------------------------------
 */

static int
PmfRenderLoop(Gp_PmfData *data, Pmf *pmf, PmfEntity *pEnt)
{
    unsigned long int i = 0;
    int stop = 0;
    int result = TCL_ERROR;
    char *varName = "gp_pmfstop";

    if (Tcl_LinkVar(data->interp, varName, 
		    (char *) &stop, TCL_LINK_BOOLEAN) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Read some vectors and render them.  If we get an end of plot
     * PMF then we are finished rendering this file.
     */

    while (*(flag *)pmf_get(pmf, NEXTV, pEnt)) {
	switch (pmfType(pEnt)) {
	    case PMF_STOP:
		result = TCL_OK;
		goto finish;
		break;
	    case PMF_LINE:
                Gp_PmfDrawLine(data, &pmfEntData(pEnt).line,
			&pmfEntOverRide(pEnt));
                break;
	    case PMF_POLYLINE:
                Gp_PmfDrawPolyLine(data, &pmfEntData(pEnt).polyline,
			&pmfEntOverRide(pEnt), pmfEntpMisc(pEnt));
                break;
	    case PMF_POLYGON:
                Gp_PmfDrawPolygon(data, &pmfEntData(pEnt).polygon,
			&pmfEntOverRide(pEnt), pmfEntpMisc(pEnt));
                break;
	    case PMF_TEXT:
                Gp_PmfDrawText(data, &pmfEntData(pEnt).text,
			&pmfEntOverRide(pEnt), pmfEntpMisc(pEnt));
                break;
	    case PMF_CIRCLE:
                Gp_PmfDrawCircle(data, &pmfEntData(pEnt).circle,
			&pmfEntOverRide(pEnt));
                break;
	    case PMF_SYMBOL:
                Gp_PmfDrawSymbol(data, &pmfEntData(pEnt).symbol,
			&pmfEntOverRide(pEnt));
                break;
	    case PMF_POLYSYMBOL:
                Gp_PmfDrawPolySymbol(data, &pmfEntData(pEnt).polysym,
			&pmfEntOverRide(pEnt), pmfEntpMisc(pEnt));
                break;
	    default:
                ASSERT(FALSE);
		goto finish;
                break;
        }
	i++;
	if (!(i%5000)) {
	    while (Tk_DoOneEvent(TK_X_EVENTS|TK_DONT_WAIT) != 0) {
		/* do nothing */
	    }
	    if (stop) {
		result = TCL_OK;
		goto finish;
	    }
	}
    }

    /*
     * Gp_Dialog(data->interp, "\"Error Drawing Pmf\"",
     *         data->interp->result, "error",
     *          "0", "Ok", (char *)NULL);
     */
  finish:
    Tcl_UnlinkVar(data->interp, varName);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfAddPosition --
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfAddPosition(Gp_PmfData *data, double dVal, int isX, double *val)
{
    CSConvert(data->cs, isX, csMILLIMETER, csPIXEL, &dVal);
    *val = dVal;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfScaleWidth --
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfScaleWidth(Gp_PmfData *data, double dVal, int *val)
{
    dVal *= data->cs->d[csMTOP];
    *val = Round(dVal);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfScaleDoubleWidth --
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfScaleDoubleWidth(Gp_PmfData *data, double dVal, double *val)
{
    *val = dVal * data->cs->d[csMTOP];
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PreprocessItem --
 *
 *	This procedure performs the first half (before the call
 *	to the items createProc) of the item "create" command. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Some memory is allocated to create the item. It will be freed
 *	when the item is deleted.
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PreprocessItem(Gp_PmfData *data, Tk_ItemType *typePtr, Tk_Item **pitem)
{
    /*
     * GP_MAINTAIN_COPY: tkCanvas.c
     */

    TkCanvas *canvasPtr = (TkCanvas *) data->canvas;
    Tk_Item *itemPtr;

    itemPtr = (Tk_Item *) ckalloc((unsigned) typePtr->itemSize);
    itemPtr->id = canvasPtr->nextId;
    canvasPtr->nextId++;
    itemPtr->tagPtr = itemPtr->staticTagSpace;
    itemPtr->tagSpace = TK_TAG_SPACE;
    itemPtr->numTags = 0;
    itemPtr->state = TK_STATE_NULL;
    itemPtr->redraw_flags = 0;
    itemPtr->updateCmd = NULL;
    itemPtr->typePtr = typePtr;
    
    itemPtr->tagPtr[0] = data->tag;
    itemPtr->numTags = 1;
    *pitem = itemPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PostprocessItem --
 *
 *	This procedure will call the bboxProc as well as perform
 *	the second half (after the call	to the items createProc)
 *	of the item "create" command.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A redraw idletask is registered.
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PostprocessItem(Tk_Canvas canvas, Tk_Item *itemPtr)
{
    TkCanvas *canvasPtr = (TkCanvas *) canvas;

    (*itemPtr->typePtr->bboxProc)(canvas, itemPtr);

    /*
     * GP_MAINTAIN_COPY: tkCanvas.c
     */

    itemPtr->nextPtr = NULL;
    canvasPtr->hotPtr = itemPtr;
    canvasPtr->hotPrevPtr = canvasPtr->lastItemPtr;
    if (canvasPtr->lastItemPtr == NULL) {
	canvasPtr->firstItemPtr = itemPtr;
    } else {
	canvasPtr->lastItemPtr->nextPtr = itemPtr;
    }
    canvasPtr->lastItemPtr = itemPtr;
    Tk_CanvasEventuallyRedraw((Tk_Canvas) canvasPtr,
		      itemPtr->x1, itemPtr->y1, itemPtr->x2, itemPtr->y2);
    canvasPtr->flags |= REPICK_NEEDED;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawLine --
 *
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawLine(data, line, override)
    Gp_PmfData *data;
    Line *line;
    Attribute *override;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    LineItem *linePtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    Tk_State state;
    unsigned long mask;
    GC newGC, arrowGC;
    Tk_Window tkwin;

    Gp_PreprocessItem(data, &tkLineType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    linePtr = (LineItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: tkCanvLine.c
     *
     * Carry out initialization that is needed to set defaults and to
     * allow proper cleanup after errors during the the remainder of
     * this procedure.
     */
 
    Tk_CreateOutline(&(linePtr->outline));
    linePtr->canvas = canvas;
    linePtr->numPoints = 0;
    linePtr->coordPtr = NULL;
    linePtr->capStyle = CapButt;
    linePtr->joinStyle = JoinRound;
    linePtr->arrowGC = None;
    if (noneUid == NULL) {
        noneUid = Tk_GetUid("none");
        firstUid = Tk_GetUid("first");
        lastUid = Tk_GetUid("last");
        bothUid = Tk_GetUid("both");
    }
    linePtr->arrow = noneUid;
    linePtr->arrowShapeA = 8.0;
    linePtr->arrowShapeB = 10.0;
    linePtr->arrowShapeC = 3.0;
    linePtr->firstArrowPtr = NULL;
    linePtr->lastArrowPtr = NULL;
    linePtr->smooth = 0;
    linePtr->splineSteps = 12;

    /*
     * Assign the positions, width and colour.
     */

    linePtr->numPoints = 2;
    linePtr->coordPtr = (double *) ckalloc((unsigned) sizeof(double) *4);

    Gp_PmfAddPosition(data, (double)PX(&lineStart(line)), 1, 
		      &(linePtr->coordPtr[0]));
    Gp_PmfAddPosition(data, (double)PY(&lineStart(line)), 0,
		      &(linePtr->coordPtr[1]));
    Gp_PmfAddPosition(data, (double)PX(&lineEnd(line)), 1,
		      &(linePtr->coordPtr[2]));
    Gp_PmfAddPosition(data, (double)PY(&lineEnd(line)), 0,
		      &(linePtr->coordPtr[3]));

    Gp_PmfScaleWidth(data, (double)attThick(override) /1000.0,
		     &(linePtr->outline.width));

    colorUid = Tk_GetUid(attStrLCol(override));
    linePtr->outline.color = Tk_GetColor(interp, tkwin, colorUid);

    /*
     * GP_MAINTAIN_COPY: tkCanvLine.c
     *
     * A few of the options require additional processing, such as
     * graphics contexts.
     */
 
    state = itemPtr->state;
 
    if(state == TK_STATE_NULL) {
        state = ((TkCanvas *)canvas)->canvas_state;
    }
 
    if (linePtr->outline.activeWidth > linePtr->outline.width ||
            linePtr->outline.activeDash.number > 0 ||
            linePtr->outline.activeColor != NULL ||
            linePtr->outline.activeStipple != None) {
        itemPtr->redraw_flags |= REDRAW_STATE_DEPENDANT;
    } else {
        itemPtr->redraw_flags &= ~REDRAW_STATE_DEPENDANT;
    }
    mask = Tk_ConfigOutlineGC(&gcValues, canvas, itemPtr,
            &(linePtr->outline));
    if (mask) {
        if (linePtr->arrow == noneUid) {
            gcValues.cap_style = linePtr->capStyle;
            mask |= GCCapStyle;
        }
        gcValues.join_style = linePtr->joinStyle;
        mask |= GCJoinStyle;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
        gcValues.line_width = 0;
        arrowGC = Tk_GetGC(tkwin, mask, &gcValues);
    } else {
        newGC = arrowGC = None;
    }
    if (linePtr->outline.gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), linePtr->outline.gc);
    }
    if (linePtr->arrowGC != None) {
        Tk_FreeGC(Tk_Display(tkwin), linePtr->arrowGC);
    }
    linePtr->outline.gc = newGC;
    linePtr->arrowGC = arrowGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawCircle --
 *
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawCircle(data, circle, override)
    Gp_PmfData *data;
    Circle *circle;
    Attribute *override;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    ArcItem *arcPtr;
    Tk_Uid colorUid;
    Tk_State state;
    XGCValues gcValues;
    GC newGC;
    unsigned long mask;
    Tk_Window tkwin;
    float flAngTwo;
    double dExtent;
    int i;
    XColor *fillColor;
    Pixmap fillStipple;

 
    Gp_PreprocessItem(data, &tkArcType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    arcPtr = (ArcItem *) itemPtr;
 
    /*
     * GP_MAINTAIN_COPY: tkCanvArc.c
     *
     * Carry out once-only initialization.
     */
 
    if (arcUid == NULL) {
        arcUid = Tk_GetUid("arc");
        chordUid = Tk_GetUid("chord");
        pieSliceUid = Tk_GetUid("pieslice");
    }
 
    /*
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */
 
    Tk_CreateOutline(&(arcPtr->outline));
    arcPtr->start = 0;
    arcPtr->extent = 90;
    arcPtr->outlinePtr = NULL;
    arcPtr->numOutlinePoints = 0;
    arcPtr->fillColor = NULL;
    arcPtr->activeFillColor = NULL;
    arcPtr->disabledFillColor = NULL;
    arcPtr->fillStipple = None;
    arcPtr->activeFillStipple = None;
    arcPtr->disabledFillStipple = None;
    arcPtr->style = pieSliceUid;
    arcPtr->fillGC = None;
 

    /*
     * Assign the positions, width, etc...
     * The value of GP_MAX_EXTENT is needed because an extent of
     * either zero or more than GP_MAX_EXTENT will cause nothing to be
     * drawn.
     */

    Gp_PmfAddPosition(data,
		(double)(PX(&circPosition(circle)) -circRadius(circle)), 1,
		&arcPtr->bbox[0]);
    Gp_PmfAddPosition(data,
		(double)(PY(&circPosition(circle)) -circRadius(circle)), 0,
		&arcPtr->bbox[1]);
    Gp_PmfAddPosition(data,
		(double)(PX(&circPosition(circle)) +circRadius(circle)), 1,
		&arcPtr->bbox[2]);
    Gp_PmfAddPosition(data,
		(double)(PY(&circPosition(circle)) +circRadius(circle)), 0,
		&arcPtr->bbox[3]);

    Gp_PmfScaleWidth(data, (double)attThick(override) /1000.0,
		     &(arcPtr->outline.width));

#define GP_MAX_EXTENT		((double)359.9999999)

    arcPtr->start = circAngleOne(circle);
    flAngTwo = circAngleTwo(circle);
    while (flAngTwo < circAngleOne(circle)) {
	flAngTwo += 360.0;
    }
    dExtent = flAngTwo - circAngleOne(circle);
    if (dExtent == 0.0 || dExtent >= GP_MAX_EXTENT) {
	dExtent = GP_MAX_EXTENT;
    }
    arcPtr->extent = dExtent;

    colorUid = Tk_GetUid(attStrLCol(override));
    if (circEnclose(circle)) {
	arcPtr->style = pieSliceUid;
	arcPtr->outline.color = Tk_GetColor(interp, tkwin, colorUid);
    } else {
	arcPtr->style = arcUid;
	arcPtr->fillColor = Tk_GetColor(interp, tkwin, colorUid);
    }

    /*
     * GP_MAINTAIN_COPY: tkCanvArc.c
     *
     * A few of the options require additional processing, such as
     * style and graphics contexts.
     */
 
    state = itemPtr->state;
 
    if (arcPtr->outline.activeWidth > arcPtr->outline.width ||
            arcPtr->outline.activeDash.number > 0 ||
            arcPtr->outline.activeColor != NULL ||
            arcPtr->outline.activeStipple != None ||
            arcPtr->activeFillColor != NULL ||
            arcPtr->activeFillStipple != None) {
        itemPtr->redraw_flags |= REDRAW_STATE_DEPENDANT;
    } else {
        itemPtr->redraw_flags &= ~REDRAW_STATE_DEPENDANT;
    }
 
    if(state == TK_STATE_NULL) {
        state = ((TkCanvas *)canvas)->canvas_state;
    }
 
    i = arcPtr->start/360.0;
    arcPtr->start -= i*360.0;
    if (arcPtr->start < 0) {
        arcPtr->start += 360.0;
    }
    i = arcPtr->extent/360.0;
    arcPtr->extent -= i*360.0;

/* 
    if ((arcPtr->style != arcUid) && (arcPtr->style != chordUid)
            && (arcPtr->style != pieSliceUid)) {
        Tcl_AppendResult(interp, "bad -style option \"",
                arcPtr->style, "\": must be arc, chord, or pieslice",
                (char *) NULL);
        arcPtr->style = pieSliceUid;
        return TCL_ERROR;
    }
*/
 
    mask = Tk_ConfigOutlineGC(&gcValues, canvas, itemPtr,
            &(arcPtr->outline));
    if (mask) {
        gcValues.cap_style = CapButt;
        mask |= GCCapStyle;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    } else {
        newGC = None;
    }
    if (arcPtr->outline.gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), arcPtr->outline.gc);
    }
    arcPtr->outline.gc = newGC;
 
    fillColor = arcPtr->fillColor;
    fillStipple = arcPtr->fillStipple;
    if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
        if (arcPtr->activeFillColor!=NULL) {
            fillColor = arcPtr->activeFillColor;
        }
        if (arcPtr->activeFillStipple!=None) {
            fillStipple = arcPtr->activeFillStipple;
        }
    } else if (state==TK_STATE_DISABLED) {
        if (arcPtr->disabledFillColor!=NULL) {
            fillColor = arcPtr->disabledFillColor;
        }
        if (arcPtr->disabledFillStipple!=None) {
            fillStipple = arcPtr->disabledFillStipple;
        }
    }
 
    if ((fillColor == NULL) || (arcPtr->style == arcUid)) {
        newGC = None;
    } else {
        gcValues.foreground = fillColor->pixel;
        if (arcPtr->style == chordUid) {
            gcValues.arc_mode = ArcChord;
        } else {
            gcValues.arc_mode = ArcPieSlice;
        }
        mask = GCForeground|GCArcMode;
        if (fillStipple != None) {
            gcValues.stipple = fillStipple;
            gcValues.fill_style = FillStippled;
            mask |= GCStipple|GCFillStyle;
        }
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (arcPtr->fillGC != None) {
        Tk_FreeGC(Tk_Display(tkwin), arcPtr->fillGC);
    }
    arcPtr->fillGC = newGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawPolyLine --
 *
 *	NOTE: the code for this is almost identical as the
 *		code for Gp_PmfDrawLine.
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawPolyLine(data, polyline, override, pvl)
    Gp_PmfData *data;
    PolyLine *polyline;
    Attribute *override;
    excVoidList *pvl;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    LineItem *linePtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    Tk_State  state;
    unsigned long mask;
    GC newGC, arrowGC;
    Tk_Window tkwin;
    int i;
    double *coordPtr;
    excPoint *ppt = vlData(pvl);

    Gp_PreprocessItem(data, &tkLineType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    linePtr = (LineItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: tkCanvLine.c
     *
     * Carry out initialization that is needed to set defaults and to
     * allow proper cleanup after errors during the the remainder of
     * this procedure.
     */
 
    Tk_CreateOutline(&(linePtr->outline));
    linePtr->canvas = canvas;
    linePtr->numPoints = 0;
    linePtr->coordPtr = NULL;
    linePtr->capStyle = CapButt;
    linePtr->joinStyle = JoinRound;
    linePtr->arrowGC = None;
    if (noneUid == NULL) {
        noneUid = Tk_GetUid("none");
        firstUid = Tk_GetUid("first");
        lastUid = Tk_GetUid("last");
        bothUid = Tk_GetUid("both");
    }
    linePtr->arrow = noneUid;
    linePtr->arrowShapeA = 8.0;
    linePtr->arrowShapeB = 10.0;
    linePtr->arrowShapeC = 3.0;
    linePtr->firstArrowPtr = NULL;
    linePtr->lastArrowPtr = NULL;
    linePtr->smooth = 0;
    linePtr->splineSteps = 12;

    /*
     * Assign the positions, width and colour.
     */

    linePtr->numPoints = vlNumElements(pvl);
    linePtr->coordPtr =
	(double *) ckalloc((unsigned) sizeof(double) *(2 *vlNumElements(pvl)));

    for (i = 0, coordPtr = linePtr->coordPtr; i < linePtr->numPoints;
					      i++, coordPtr += 2) {
        Gp_PmfAddPosition(data, (double)PX(&(ppt[i])), 1,
			  &(coordPtr[0]));
        Gp_PmfAddPosition(data, (double)PY(&(ppt[i])), 0,
			  &(coordPtr[1]));
    }
 
    Gp_PmfScaleWidth(data, (double)attThick(override) /1000.0,
		     &(linePtr->outline.width));

    colorUid = Tk_GetUid(attStrLCol(override));
    linePtr->outline.color = Tk_GetColor(interp, tkwin, colorUid);

    /*
     * GP_MAINTAIN_COPY: tkCanvLine.c
     *
     * A few of the options require additional processing, such as
     * graphics contexts.
     */
 
    state = itemPtr->state;
 
    if(state == TK_STATE_NULL) {
        state = ((TkCanvas *)canvas)->canvas_state;
    }
 
    if (linePtr->outline.activeWidth > linePtr->outline.width ||
            linePtr->outline.activeDash.number > 0 ||
            linePtr->outline.activeColor != NULL ||
            linePtr->outline.activeStipple != None) {
        itemPtr->redraw_flags |= REDRAW_STATE_DEPENDANT;
    } else {
        itemPtr->redraw_flags &= ~REDRAW_STATE_DEPENDANT;
    }
    mask = Tk_ConfigOutlineGC(&gcValues, canvas, itemPtr,
            &(linePtr->outline));
    if (mask) {
        if (linePtr->arrow == noneUid) {
            gcValues.cap_style = linePtr->capStyle;
            mask |= GCCapStyle;
        }
        gcValues.join_style = linePtr->joinStyle;
        mask |= GCJoinStyle;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
        gcValues.line_width = 0;
        arrowGC = Tk_GetGC(tkwin, mask, &gcValues);
    } else {
        newGC = arrowGC = None;
    }
    if (linePtr->outline.gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), linePtr->outline.gc);
    }
    if (linePtr->arrowGC != None) {
        Tk_FreeGC(Tk_Display(tkwin), linePtr->arrowGC);
    }
    linePtr->outline.gc = newGC;
    linePtr->arrowGC = arrowGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawPolygon --
 *
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawPolygon(data, polygon, override, pvl)
    Gp_PmfData *data;
    Polygon *polygon;
    Attribute *override;
    excVoidList *pvl;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    PolygonItem *polyPtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    Tk_State state;
    unsigned long mask;
    GC newGC;
    Tk_Window tkwin;
    int i;
    double *coordPtr;
    excPoint *ppt = vlData(pvl);
    XColor *fillColor;
    Pixmap fillStipple;

    Gp_PreprocessItem(data, &tkPolygonType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    polyPtr = (PolygonItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: tkCanvPoly.c
     *
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */
 
    Tk_CreateOutline(&(polyPtr->outline));
    polyPtr->numPoints = 0;
    polyPtr->pointsAllocated = 0;
    polyPtr->coordPtr = NULL;
    polyPtr->joinStyle = JoinRound;
    polyPtr->fillColor = NULL;
    polyPtr->activeFillColor = NULL;
    polyPtr->disabledFillColor = NULL;
    polyPtr->fillStipple = None;
    polyPtr->activeFillStipple = None;
    polyPtr->disabledFillStipple = None;
    polyPtr->fillGC = None;
    polyPtr->smooth = 0;
    polyPtr->splineSteps = 12;
 
    /*
     * Assign the positions, width and colour.
     */

    polyPtr->numPoints = vlNumElements(pvl);
    polyPtr->coordPtr =	(double *) ckalloc((unsigned) 
			   sizeof(double) *(2 *(vlNumElements(pvl) +1)));

    for (i = 0, coordPtr = polyPtr->coordPtr; i < polyPtr->numPoints;
					      i++, coordPtr += 2) {
        Gp_PmfAddPosition(data, (double)PX(&(ppt[i])), 1,
			  &(coordPtr[0]));
        Gp_PmfAddPosition(data, (double)PY(&(ppt[i])), 0,
			  &(coordPtr[1]));
    }
    polyPtr->numPoints++;
    coordPtr[0] = polyPtr->coordPtr[0];
    coordPtr[1] = polyPtr->coordPtr[1];

    Gp_PmfScaleWidth(data, (double)attThick(override) /1000.0,
		     &(polyPtr->outline.width));

    colorUid = Tk_GetUid(attStrLCol(override));
    polyPtr->outline.color = Tk_GetColor(interp, tkwin, colorUid);
    colorUid = Tk_GetUid(attStrFCol(override));
    polyPtr->fillColor = Tk_GetColor(interp, tkwin, colorUid);

    /*
     * GP_MAINTAIN_COPY: tkCanvPoly.c
     *
     * A few of the options require additional processing, such as
     * graphics contexts.
     */
 
    state = itemPtr->state;
 
    if (polyPtr->outline.activeWidth > polyPtr->outline.width ||
            polyPtr->outline.activeDash.number > 0 ||
            polyPtr->outline.activeColor != NULL ||
            polyPtr->outline.activeStipple != None ||
            polyPtr->activeFillColor != NULL ||
            polyPtr->activeFillStipple != None) {
        itemPtr->redraw_flags |= REDRAW_STATE_DEPENDANT;
    } else {
        itemPtr->redraw_flags &= ~REDRAW_STATE_DEPENDANT;
    }
 
    if(state == TK_STATE_NULL) {
        state = ((TkCanvas *)canvas)->canvas_state;
    }
    if (state==TK_STATE_HIDDEN) {
	goto end;
/*
        ComputePolygonBbox(canvas, (Tk_Item *) polyPtr);
        return TCL_OK;
*/
    }
 
    mask = Tk_ConfigOutlineGC(&gcValues, canvas, itemPtr, &(polyPtr->outline));
    if (mask) {
        gcValues.cap_style = CapRound;
        gcValues.join_style = polyPtr->joinStyle;
        mask |= GCCapStyle|GCJoinStyle;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    } else {
        newGC = None;
    }
    if (polyPtr->outline.gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), polyPtr->outline.gc);
    }
    polyPtr->outline.gc = newGC;
 
    fillColor = polyPtr->fillColor;
    fillStipple = polyPtr->fillStipple;
    if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
        if (polyPtr->activeFillColor!=NULL) {
            fillColor = polyPtr->activeFillColor;
        }
        if (polyPtr->activeFillStipple!=None) {
            fillStipple = polyPtr->activeFillStipple;
        }
    } else if (state==TK_STATE_DISABLED) {
        if (polyPtr->disabledFillColor!=NULL) {
            fillColor = polyPtr->disabledFillColor;
        }
        if (polyPtr->disabledFillStipple!=None) {
            fillStipple = polyPtr->disabledFillStipple;
        }
    }
 
 
    if (fillColor == NULL) {
        newGC = None;
    } else {
        gcValues.foreground = fillColor->pixel;
        mask = GCForeground;
        if (fillStipple != None) {
            gcValues.stipple = fillStipple;
            gcValues.fill_style = FillStippled;
            mask |= GCStipple|GCFillStyle;
        }
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (polyPtr->fillGC != None) {
        Tk_FreeGC(Tk_Display(tkwin), polyPtr->fillGC);
    }
    polyPtr->fillGC = newGC;
 
    /*
     * Finally do the postprocessing.
     */

  end:
    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawText --
 *
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawText(data, text, override, pvl)
    Gp_PmfData *data;
    Text *text;
    Attribute *override;
    excVoidList *pvl;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    GpTextItem *textPtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    char *str;
    GC newGC;
    unsigned long mask;
    Tk_Window tkwin;

    Gp_PreprocessItem(data, &GpTextType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    textPtr = (GpTextItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: gpCanvText.c
     *
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */
 
    textPtr->text = NULL;
    textPtr->numChars = 0;
    textPtr->x = 0;
    textPtr->y = 0;
    textPtr->rotation = 0;
    textPtr->outlineWidth = 0;
    textPtr->colorOutline = NULL;
    textPtr->charWidth = 10;
    textPtr->charHeight = 10;
    textPtr->charSpace = 1;
    textPtr->charSlant = 0;
    textPtr->justify = TK_JUSTIFY_LEFT;
    textPtr->pfont = NULL;
    textPtr->gc = None;
 
    textPtr->geom.flXOrigin = 0;
    textPtr->geom.flYOrigin = 0;
    textPtr->geom.flXScale = textPtr->charWidth/(double) GPFONT_SRES;
    textPtr->geom.flYScale = textPtr->charHeight/(double) GPFONT_SRES;
    textPtr->geom.flCos = 1;
    textPtr->geom.flSin = 0;
    textPtr->geom.flSlant = 0;
    textPtr->geom.flCharSpace = textPtr->charSpace;
    textPtr->geom.flLength = (float) textPtr->charWidth;
 
    /*
     * Assign the positions, width and colour, etc...
     */

    Gp_PmfAddPosition(data, (double)PX(&textPointOne(text)), 1,
		      &(textPtr->x));
    Gp_PmfAddPosition(data, (double)PY(&textPointOne(text)), 0,
		      &(textPtr->y));

    Gp_PmfScaleDoubleWidth(data, (double)attThick(override) /1000.0,
			   &(textPtr->outlineWidth));
 
    Gp_PmfScaleDoubleWidth(data, (double)attText(override).flWd,
			   &(textPtr->charWidth));
    Gp_PmfScaleDoubleWidth(data, (double)attText(override).flHt,
			   &(textPtr->charHeight));
    Gp_PmfScaleDoubleWidth(data, (double)attText(override).flSp,
			   &(textPtr->charSpace));

    textPtr->charSlant = attText(override).flSl;

    colorUid = Tk_GetUid(attStrLCol(override));
    textPtr->colorOutline = Tk_GetColor(interp, tkwin, colorUid);

    textPtr->pfont = Gp_GetFont(interp, attFont(override));

    switch (textPlacement(text)) {
	case TCENTER:
	case TTOTALCENTER:
	    textPtr->justify = TK_JUSTIFY_CENTER;
	    break;
	case TRIGHT:
	    textPtr->justify = TK_JUSTIFY_RIGHT;
	    break;
	case TLEFT:
	default:
	    textPtr->justify = TK_JUSTIFY_LEFT;
	    break;
    }
    switch (textPlacement(text)) {
	case TCENTER:
	case TTOTALCENTER:
	case TRIGHT:
	case TLEFT:
	    textPtr->rotation = PX(&textPointTwo(text));
	    break;
	default:
	    textPtr->rotation = 0.0;
	    break;
    }

    str = (char *) vlData(pvl);
    textPtr->text = (char *) ckalloc((unsigned) (strlen(str) +1));
    strcpy(textPtr->text, str);

    /*
     * GP_MAINTAIN_COPY: gpCanvText.c
     *
     * A few of the options require additional processing, such as
     * graphics contexts.
     */
 
    newGC = None;
    if (textPtr->colorOutline != NULL) {
        if (textPtr->outlineWidth < 0) {
            textPtr->outlineWidth = 1;
        }
        gcValues.foreground = textPtr->colorOutline->pixel;
        gcValues.line_width = textPtr->outlineWidth;
        mask = GCForeground|GCLineWidth;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (textPtr->gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), textPtr->gc);
    }
    textPtr->gc = newGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawSymbol --
 *
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawSymbol(data, symbol, override)
    Gp_PmfData *data;
    Symbol *symbol;
    Attribute *override;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    GpSymbolItem *symbolPtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    GC newGC;
    unsigned long mask;
    Tk_Window tkwin;

    Gp_PreprocessItem(data, &GpSymbolType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    symbolPtr = (GpSymbolItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: gpCanvSymb.c
     *
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */
 
    symbolPtr->symbol = 1;
    symbolPtr->numPoints = 0;
    symbolPtr->coordPtr = NULL;
    symbolPtr->rotation = 0;
    symbolPtr->outlineWidth = 1;
    symbolPtr->colorOutline = NULL;
    symbolPtr->charWidth = 10;
    symbolPtr->charHeight = 10;
    symbolPtr->charSpace = 1;
    symbolPtr->charSlant = 0;
    symbolPtr->pfont = NULL;
    symbolPtr->gc = None;
 
    symbolPtr->geom.flXOrigin = 0;
    symbolPtr->geom.flYOrigin = 0;
    symbolPtr->geom.flXScale = symbolPtr->charWidth/(double) GPFONT_SRES;
    symbolPtr->geom.flYScale = symbolPtr->charHeight/(double) GPFONT_SRES;
    symbolPtr->geom.flCos = 1;
    symbolPtr->geom.flSin = 0;
    symbolPtr->geom.flSlant = 0;
    symbolPtr->geom.flCharSpace = symbolPtr->charSpace;
    symbolPtr->geom.flLength = (float) symbolPtr->charWidth;
 
    /*
     * Assign the positions, width and colour, etc...
     */

    symbolPtr->symbol = symbSymbol(symbol); 
    symbolPtr->numPoints = 1;
    symbolPtr->coordPtr = (double *) ckalloc((unsigned) sizeof(double) *2);
 
    Gp_PmfAddPosition(data, (double)PX(&symbPosition(symbol)), 1,
                      &(symbolPtr->coordPtr[0]));
    Gp_PmfAddPosition(data, (double)PY(&symbPosition(symbol)), 0,
                      &(symbolPtr->coordPtr[1]));

    Gp_PmfScaleDoubleWidth(data, (double)attThick(override) /1000.0,
			   &(symbolPtr->outlineWidth));
 
    Gp_PmfScaleDoubleWidth(data,
			   (double)attText(override).flWd *symbSize(symbol),
			   &(symbolPtr->charWidth));
    Gp_PmfScaleDoubleWidth(data,
			   (double)attText(override).flHt *symbSize(symbol),
			   &(symbolPtr->charHeight));
    Gp_PmfScaleDoubleWidth(data,
			   (double)attText(override).flSp *symbSize(symbol),
			   &(symbolPtr->charSpace));

    symbolPtr->charSlant = attText(override).flSl;
    symbolPtr->rotation = symbAngle(symbol);

    colorUid = Tk_GetUid(attStrLCol(override));
    symbolPtr->colorOutline = Tk_GetColor(interp, tkwin, colorUid);

    symbolPtr->pfont = Gp_GetFont(interp, "stroke.tbl");

    /*
     * GP_MAINTAIN_COPY: gpCanvSymb.c
     *
     * A few of the options require additional processing, such as
     * graphics consymbols.
     */
 
    newGC = None;
    if (symbolPtr->colorOutline != NULL) {
        if (symbolPtr->outlineWidth < 0) {
            symbolPtr->outlineWidth = 1;
        }
        gcValues.foreground = symbolPtr->colorOutline->pixel;
        gcValues.line_width = symbolPtr->outlineWidth;
        mask = GCForeground|GCLineWidth;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (symbolPtr->gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), symbolPtr->gc);
    }
    symbolPtr->gc = newGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PmfDrawPolySymbol --
 *
 *	NOTE: the code for this is almost identical as the
 *		code for Gp_PmfDrawSymbol.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
Gp_PmfDrawPolySymbol(data, polysymbol, override, pvl)
    Gp_PmfData *data;
    PolySym *polysymbol;
    Attribute *override;
    excVoidList *pvl;
{
    Tcl_Interp *interp = data->interp;
    Tk_Canvas canvas = data->canvas;
    Tk_Item *itemPtr;
    GpSymbolItem *symbolPtr;
    XGCValues gcValues;
    Tk_Uid colorUid;
    GC newGC;
    unsigned long mask;
    Tk_Window tkwin;
    int i;
    double *coordPtr;
    excPoint *ppt = vlData(pvl);

    Gp_PreprocessItem(data, &GpSymbolType, &itemPtr);
    tkwin = Tk_CanvasTkwin(canvas);
    symbolPtr = (GpSymbolItem *) itemPtr;

    /*
     * GP_MAINTAIN_COPY: gpCanvSymb.c
     *
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */
 
    symbolPtr->symbol = 1;
    symbolPtr->numPoints = 0;
    symbolPtr->coordPtr = NULL;
    symbolPtr->rotation = 0;
    symbolPtr->outlineWidth = 1;
    symbolPtr->colorOutline = NULL;
    symbolPtr->charWidth = 10;
    symbolPtr->charHeight = 10;
    symbolPtr->charSpace = 1;
    symbolPtr->charSlant = 0;
    symbolPtr->pfont = NULL;
    symbolPtr->gc = None;
 
    symbolPtr->geom.flXOrigin = 0;
    symbolPtr->geom.flYOrigin = 0;
    symbolPtr->geom.flXScale = symbolPtr->charWidth/(double) GPFONT_SRES;
    symbolPtr->geom.flYScale = symbolPtr->charHeight/(double) GPFONT_SRES;
    symbolPtr->geom.flCos = 1;
    symbolPtr->geom.flSin = 0;
    symbolPtr->geom.flSlant = 0;
    symbolPtr->geom.flCharSpace = symbolPtr->charSpace;
    symbolPtr->geom.flLength = (float) symbolPtr->charWidth;
 
    /*
     * Assign the positions, width and colour.
     */

    symbolPtr->symbol = psymSymbol(polysymbol); 
    symbolPtr->numPoints = vlNumElements(pvl);
    symbolPtr->coordPtr =
	(double *) ckalloc((unsigned) sizeof(double) *(2 *vlNumElements(pvl)));

    for (i = 0, coordPtr = symbolPtr->coordPtr; i < symbolPtr->numPoints;
					      i++, coordPtr += 2) {
        Gp_PmfAddPosition(data, (double)PX(&(ppt[i])), 1,
			  &(coordPtr[0]));
        Gp_PmfAddPosition(data, (double)PY(&(ppt[i])), 0,
			  &(coordPtr[1]));
    }

    Gp_PmfScaleDoubleWidth(data, (double)attThick(override) /1000.0,
			   &(symbolPtr->outlineWidth));
 
    Gp_PmfScaleDoubleWidth(data,
		   (double)attText(override).flWd *psymSize(polysymbol),
		   &(symbolPtr->charWidth));
    Gp_PmfScaleDoubleWidth(data,
		   (double)attText(override).flHt *psymSize(polysymbol),
		   &(symbolPtr->charHeight));
    Gp_PmfScaleDoubleWidth(data,
		   (double)attText(override).flSp *psymSize(polysymbol),
		   &(symbolPtr->charSpace));

    symbolPtr->charSlant = attText(override).flSl;
    symbolPtr->rotation = psymAngle(polysymbol);

    colorUid = Tk_GetUid(attStrLCol(override));
    symbolPtr->colorOutline = Tk_GetColor(interp, tkwin, colorUid);

    symbolPtr->pfont = Gp_GetFont(interp, "stroke.tbl");

    /*
     * GP_MAINTAIN_COPY: gpCanvSymb.c
     *
     * A few of the options require additional processing, such as
     * graphics consymbols.
     */
 
    newGC = None;
    if (symbolPtr->colorOutline != NULL) {
        if (symbolPtr->outlineWidth < 0) {
            symbolPtr->outlineWidth = 1;
        }
        gcValues.foreground = symbolPtr->colorOutline->pixel;
        gcValues.line_width = symbolPtr->outlineWidth;
        mask = GCForeground|GCLineWidth;
        newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (symbolPtr->gc != None) {
        Tk_FreeGC(Tk_Display(tkwin), symbolPtr->gc);
    }
    symbolPtr->gc = newGC;
 
    /*
     * Finally do the postprocessing.
     */

    Gp_PostprocessItem(canvas, itemPtr);

}
